/*

cc -g sam_comp1.c -DHAVE_LIBLZMA -I/software/badger/opt/xz_utils/include -L/software/badger/opt/xz_utils/lib -lz -llzma

See end for sizes

 */

#define SIMPLE_READGROUP_MODE
//#define SIMPLE_QNAME_MODE
#define SIMPLE_CIGAR_MODE
#define SIMPLE_SEQ_MODE
#define SIMPLE_QUAL_MODE
//#define SIMPLE_AUX_MODE

#define USE_RLE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <zlib.h>
#include <assert.h>
#include <stdarg.h>
#include <malloc.h>
#include <fcntl.h>

#define COMP_MODE_NONE 0
#define COMP_MODE_ZLIB 1
#define COMP_MODE_LZMA 2

/* ------------------------------------------------------------------------ */
/* Common code, to move somewhere else */

#ifdef USE_RLE
static char rle_last = 0;
static int  rle_rep = 0;

void rle_init(void) {
    rle_last = 0;
    rle_rep = 0;
}

int u72int(unsigned char *u7, uint32_t *out);
static char rle_get(char **cpp) {
    char c;
    unsigned char *cp = (unsigned char *)*cpp;

    //printf("cp=%d,%d rle_rep=%d last=%d ", cp[0],cp[1], rle_rep, rle_last);

    if (rle_rep) {
	rle_rep--;
	c = rle_last;
    } else {
	c = *cp++;
	if (c == rle_last) {
	    cp += u72int(cp, &rle_rep);
	} else {
	    rle_last = c;
	}
    }

    //printf("=>%c and cp +%d\n", c, (char *)cp-(char *)*cpp);

    *cpp = (char *)cp;
    return c;
}

#else

/* Dummy rle that does nothing */
#define rle_init()
#define rle_get(cpp) ((*cpp)++,(*cpp)[-1])
#endif

#define ABS(i) ((i)>0?(i):-(i))
#define MIN(a,b) ((a)<(b)?(a):(b))
#define MAX(a,b) ((a)>(b)?(a):(b))

char *append_int(char *cp, int i) {
    int j, k = 0;

    if (i < 0) {
	*cp++ = '-';
	i = -i;
    } else if (i == 0) {
	*cp++ = '0';
	return cp;
    }

    if (i < 1000)
	goto b1;
    if (i < 100000)
	goto b2;
    if (i < 100000000)
	goto b3;

    j = i / 1000000000;
    if (j || k) *cp++ = j + '0', k=1, i %= 1000000000;

    j = i / 100000000;
    if (j || k) *cp++ = j + '0', k=1, i %= 100000000;
    
 b3:
    j = i / 10000000;
    if (j || k) *cp++ = j + '0', k=1, i %= 10000000;
    
    j = i / 1000000;
    if (j || k) *cp++ = j + '0', k=1, i %= 1000000;
    
    j = i / 100000;
    if (j || k) *cp++ = j + '0', k=1, i %= 100000;
    
 b2:
    j = i / 10000;
    if (j || k) *cp++ = j + '0', k=1, i %= 10000;

    j = i / 1000;
    if (j || k) *cp++ = j + '0', k=1, i %= 1000;

 b1:
    j = i / 100;
    if (j || k) *cp++ = j + '0', k=1, i %= 100;

    j = i / 10;
    if (j || k) *cp++ = j + '0', k=1, i %= 10;

    if (i || k) *cp++ = i + '0';

    return cp;
}

#define EVEN_SIMPLER
int fast_fprintf(FILE *fp, char *fmt, ...) {
    char *cp;
    long l;
    int i;
    double d; 
    char tmp[128], *ct;
    va_list ap;

    va_start(ap, fmt);

    for (cp = fmt; *cp; cp++) {
	switch(*cp) {

	/* A format specifier */
	case '%': {
	    long conv_len1=0, conv_len2=0, conv_len=0;
	    signed int arg_size;

	    cp++;
#ifndef EVEN_SIMPLER
	    {
		char c;

		/* Firstly, strip the modifier flags (+-#0 and [space]) */
		for(; c=*cp;) {
		    if ('#' == c)
			;
		    else if ('-' == c || '+' == c || ' ' == c)
			;
		    else
			break;
		}
	    }

	    /* Width specifier */
	    if (*cp >= '0' && *cp <= '9') {
		for (l = 0; *cp >= '0' && *cp <= '9'; cp++)
		    l = l*10 + *cp - '0';
		conv_len = conv_len1 = l;
	    } else if (*cp == '*') {
		conv_len = conv_len1 = (int)va_arg(ap, int);
		cp++;
	    }
#endif

	    /* Precision specifier */
	    if ('.' == *cp) {
		cp++;
		for (l = 0; *cp >= '0' && *cp <= '9'; cp++)
		    l = l*10 + *cp - '0';
		conv_len2 = l;
		if (*cp == '*') {
		    conv_len2 = (int)va_arg(ap, int);
		    cp++;
		}
		conv_len = MAX(conv_len1, conv_len2);
	    }

#ifndef EVEN_SIMPLER
	    /* Short/long identifier */
	    if ('h' == *cp) {
		arg_size = -1; /* short */
		cp++;
	    } else if ('l' == *cp) {
		arg_size = 1; /* long */
		cp++;
	    } else {
		arg_size = 0; /* int */
	    }
#else
	    arg_size = 0;
#endif

	    /* The actual type */
	    switch (*cp) {
	    case '%':
		/*
		 * Not real ANSI I suspect, but we'll allow for the
		 * completely daft "%10%" example.
		 */
		putc('%', fp);
		break;

	    case 'd':
	    case 'i':
	    case 'u':
		/* Remember: char and short are sent as int on the stack */
		if (arg_size == -1)
		    l = (long)va_arg(ap, int);
		else if (arg_size == 1)
		    l = va_arg(ap, long); 
		else 
		    l = (long)va_arg(ap, int);

		ct = append_int(tmp, (int)l);
		*ct++ = 0;
		fputs(tmp, fp);
		break;

	    case 'c':
		i = va_arg(ap, int);
		putc(i, fp);
		break;

	    case 'f':
		d = va_arg(ap, double);
		fprintf(fp, "%f", d);
		break;

	    case 'e':
	    case 'E':
	    case 'g':
	    case 'G':
		d = va_arg(ap, double);
		fprintf(fp, "%g", d);
		break;

	    case 'p':
	    case 'x':
	    case 'X':
		l = (long)va_arg(ap, void *);
		puts("TODO");
		break;

	    case 'n':
		/* produces no output */
		break;

	    case 's': {
		char *s = (char *)va_arg(ap, char *);
		if (conv_len2) {
		    fwrite(s, conv_len2, 1, fp);
		} else
		    fputs(s, fp);
		break;
	    }

	    default:
		/* wchar_t types of 'C' and 'S' aren't supported */
		puts("TODO");
	    }
	    
	}

	case '\0':
	    break;

	default:
	    putc(*cp, fp);
	}
    }

    va_end(ap);
    
    return 0;
}

/* Max 8k of output */
int faster_fprintf(FILE *fp, char *fmt, ...) {
    char *cp;
    long l;
    int i;
    double d; 
    va_list ap;
    char max_line[8192], *out = max_line;

    va_start(ap, fmt);

    for (cp = fmt; *cp; cp++) {
	switch(*cp) {

	/* A format specifier */
	case '%': {
	    long conv_len1=0, conv_len2=0, conv_len=0;
	    signed int arg_size;

	    cp++;

#ifndef EVEN_SIMPLER
	    {
		char c;

		/* Firstly, strip the modifier flags (+-#0 and [space]) */
		for(; c=*cp;) {
		    if ('#' == c)
			;
		    else if ('-' == c || '+' == c || ' ' == c)
			;
		    else
			break;
		}
	    }

	    /* Width specifier */
	    if (*cp >= '0' && *cp <= '9') {
		for (l = 0; *cp >= '0' && *cp <= '9'; cp++)
		    l = l*10 + *cp - '0';
		conv_len = conv_len1 = l;
	    } else if (*cp == '*') {
		conv_len = conv_len1 = (int)va_arg(ap, int);
		cp++;
	    }
#endif

	    /* Precision specifier */
	    if ('.' == *cp) {
		cp++;
		for (l = 0; *cp >= '0' && *cp <= '9'; cp++)
		    l = l*10 + *cp - '0';
		conv_len2 = l;
		if (*cp == '*') {
		    conv_len2 = (int)va_arg(ap, int);
		    cp++;
		}
		conv_len = MAX(conv_len1, conv_len2);
	    }

#ifndef EVEN_SIMPLER
	    /* Short/long identifier */
	    if ('h' == *cp) {
		arg_size = -1; /* short */
		cp++;
	    } else if ('l' == *cp) {
		arg_size = 1; /* long */
		cp++;
	    } else {
		arg_size = 0; /* int */
	    }
#else
	    arg_size = 0;
#endif

	    /* The actual type */
	    switch (*cp) {
	    case '%':
		/*
		 * Not real ANSI I suspect, but we'll allow for the
		 * completely daft "%10%" example.
		 */
		*out++ = '%';
		break;

	    case 'd':
	    case 'i':
	    case 'u':
		/* Remember: char and short are sent as int on the stack */
		if (arg_size == -1)
		    l = (long)va_arg(ap, int);
		else if (arg_size == 1)
		    l = va_arg(ap, long); 
		else 
		    l = (long)va_arg(ap, int);

		out = append_int(out, (int)l);
		break;

	    case 'c':
		i = va_arg(ap, int);
		*out++ = i;
		break;

	    case 'f':
		d = va_arg(ap, double);
		out += sprintf(out, "%f", d);
		break;

	    case 'e':
	    case 'E':
	    case 'g':
	    case 'G':
		d = va_arg(ap, double);
		out += sprintf(out, "%g", d);
		break;

	    case 'p':
	    case 'x':
	    case 'X':
		l = (long)va_arg(ap, void *);
		puts("TODO");
		break;

	    case 'n':
		/* produces no output */
		break;

	    case 's': {
		char *s = (char *)va_arg(ap, char *);
		if (conv_len2) {
		    //memcpy(out, s, conv_len2);
		    //out += conv_len2;

		    //strncpy(out, s, conv_len2);
		    //out += MIN(conv_len2, strnlen(s));

		    while(conv_len2 && *s) {
		        *out++ = *s++;
		        conv_len2--;
		    }
		} else {
		    //strcpy(out, s);
		    //out += strlen(s);

		    while(*s) *out++ = *s++;
		}
		//		if (conv_len2) {
		//		    fwrite(s, conv_len2, 1, fp);
		//		} else
		//		    fputs(s, fp);
		break;
	    }

	    default:
		/* wchar_t types of 'C' and 'S' aren't supported */
		puts("TODO");
	    }
	    
	}

	case '\0':
	    break;

	default:
	    *out++ = *cp;
	}
    }

    *out = 0;
    fwrite(max_line, 1, out-max_line, fp);

    va_end(ap);
    
    return 0;
}

/*
 * Dummy nul-compression functions. These do nothing except satisfy
 * the internal API requirements. We use them simply to compute the
 * base-line so we can estimate the overhead due to zlib vs overhead
 * due to lzma.
 */
static char *nul_mem_deflate(char *data, size_t size, size_t *cdata_size) {
    char *out = malloc(size);
    memcpy(out, data, size);
    *cdata_size = size;
    return out;
}

static char *nul_mem_deflate_parts(char *data,
				   size_t *part_size, int nparts,
				   size_t *cdata_size) {
    size_t tot_size = 0;
    int i;
    for (i = 0; i < nparts; i++)
	tot_size += part_size[i];
    return nul_mem_deflate(data, tot_size, cdata_size);
}

static char *nul_mem_deflate_lparts(char *data,
				    size_t *part_size, int *level, int nparts,
				    size_t *cdata_size) {
    return nul_mem_deflate_parts(data, part_size, nparts, cdata_size);
}

static char *nul_mem_inflate(char *cdata, size_t csize, size_t *size) {
    char *out = malloc(csize);
    memcpy(out, cdata, csize);
    *size = csize;
    return out;
}

#ifdef HAVE_LIBLZMA
/* ------------------------------------------------------------------------ */
/*
 * Data compression routines using liblzma (xz)
 *
 * On a test set this shrunk the main db from 136157104 bytes to 114796168, but
 * caused tg_index to grow from 2m43.707s to 15m3.961s. Exporting as bfastq
 * went from 18.3s to 36.3s. So decompression suffers too, but not as bad
 * as compression times.
 *
 * For now we disable this functionality. If it's to be reenabled make sure you
 * improve the mem_inflate implementation as it's just a test hack at the
 * moment.
 */
#include <lzma.h>

/* Fast mode, but not too bad on compression still */
#define LZMA_LEVEL 3

static char *lzma_mem_deflate(char *data, size_t size, size_t *cdata_size) {
    char *out;
    size_t out_size = lzma_stream_buffer_bound(size);
    *cdata_size = 0;

    out = malloc(out_size);

    /* Single call compression */
    if (LZMA_OK != lzma_easy_buffer_encode(LZMA_LEVEL,
					   LZMA_CHECK_CRC32, NULL,
					   (uint8_t *)data, size,
					   (uint8_t *)out, cdata_size,
					   out_size))
    	return NULL;

    return out;
}

static char *lzma_mem_deflate_parts(char *data,
				    size_t *part_size, int nparts,
				    size_t *cdata_size) {
    size_t tot_size = 0;
    int i;
#if 0
    for (i = 0; i < nparts; i++)
	tot_size += part_size[i];
    return lzma_mem_deflate(data, tot_size, cdata_size);

#else
    char  *lzo[100], *out;
    size_t lzs[100];

    /* Compress parts */
    for (i = 0; i < nparts; i++) {
	lzo[i] = lzma_mem_deflate(data, part_size[i], &lzs[i]);
	fprintf(stderr, "part %2d, size %7ld, comp_size %7ld, fract %f\n",
		i, part_size[i], lzs[i], (100.0 * lzs[i]) / part_size[i]);
	data += part_size[i];
	tot_size += lzs[i];
    }

    /* Stitch together */
    data = out = malloc(tot_size);
    fprintf(stderr, "=> total size %ld\n", tot_size);
    for (i = 0; i < nparts; i++) {
	memcpy(data, lzo[i], lzs[i]);
	free(lzo[i]);
	data += lzs[i];
    }

    *cdata_size = tot_size;
    return out;
#endif
}

static char *lzma_mem_deflate_lparts(char *data,
				     size_t *part_size, int *level, int nparts,
				     size_t *cdata_size) {
    return lzma_mem_deflate_parts(data, part_size, nparts, cdata_size);
}

static char *lzma_mem_inflate(char *cdata, size_t csize, size_t *size) {
    lzma_stream strm = LZMA_STREAM_INIT;
    size_t out_size = 0, out_pos = 0;
    char *out = NULL;
    int r;

    /* Initiate the decoder */
    if (LZMA_OK != lzma_stream_decoder(&strm, 100000000, 0))
	return NULL;

    /* Decode loop */
    strm.avail_in = csize;
    strm.next_in = (uint8_t *)cdata;

    for (;strm.avail_in;) {
	if (strm.avail_in > out_size - out_pos) {
	    out_size += strm.avail_in * 4 + 32768;
	    out = realloc(out, out_size);
	}
	strm.avail_out = out_size - out_pos;
	strm.next_out = (uint8_t *)&out[out_pos];

	r = lzma_code(&strm, LZMA_RUN);
	if (LZMA_OK != r && LZMA_STREAM_END != r) {
	    fprintf(stderr, "r=%d\n", r);
	    fprintf(stderr, "mem=%ld\n", lzma_memusage(&strm));
	    return NULL;
	}

	out_pos = strm.total_out;

	if (r == LZMA_STREAM_END)
	    break;
    }

    /* finish up any unflushed data; necessary? */
    r = lzma_code(&strm, LZMA_FINISH);
    if (r != LZMA_OK && r != LZMA_STREAM_END) {
	fprintf(stderr, "r=%d\n", r);
	return NULL;
    }

    out = realloc(out, strm.total_out);
    *size = strm.total_out;

    lzma_end(&strm);

    return out;
}
#endif

/* ------------------------------------------------------------------------ */
/*
 * Data compression routines using zlib.
 */
#include <zlib.h>
static char *zlib_mem_deflate(char *data, size_t size, size_t *cdata_size) {
    z_stream s;
    unsigned char *cdata = NULL; /* Compressed output */
    int cdata_alloc = 0;
    int cdata_pos = 0;
    int err;

    cdata = malloc(cdata_alloc = size*1.05+10);
    cdata_pos = 0;

    /* Initialise zlib stream */
    s.zalloc = Z_NULL; /* use default allocation functions */
    s.zfree  = Z_NULL;
    s.opaque = Z_NULL;
    s.next_in  = (unsigned char *)data;
    s.avail_in = size;
    s.total_in = 0;
    s.next_out  = cdata;
    s.avail_out = cdata_alloc;
    s.total_out = 0;
    s.data_type = Z_BINARY;

    err = deflateInit2(&s, Z_DEFAULT_COMPRESSION, Z_DEFLATED, 15,
		       9, Z_DEFAULT_STRATEGY);

    /* Encode to 'cdata' array */
    for (;s.avail_in;) {
	s.next_out = &cdata[cdata_pos];
	s.avail_out = cdata_alloc - cdata_pos;
	if (cdata_alloc - cdata_pos <= 0) {
	    fprintf(stderr, "Deflate produced larger output than expected. Abort\n"); 
	    return NULL;
	}
	err = deflate(&s, Z_NO_FLUSH);
	cdata_pos = cdata_alloc - s.avail_out;
	if (err != Z_OK) {
	    fprintf(stderr, "zlib deflate error: %s\n", s.msg);
	    break;
	}
    }
    if (deflate(&s, Z_FINISH) != Z_STREAM_END) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    *cdata_size = s.total_out;

    if (deflateEnd(&s) != Z_OK) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    return (char *)cdata;
}

static char *zlib_mem_deflate_parts(char *data,
				    size_t *part_size, int nparts,
				    size_t *cdata_size) {
    z_stream s;
    unsigned char *cdata = NULL; /* Compressed output */
    int cdata_alloc = 0;
    int cdata_pos = 0;
    int err, i;
    size_t size = 0;

    for (i = 0; i < nparts; i++)
	size += part_size[i];

    cdata = malloc(cdata_alloc = size*1.05+256*nparts);
    cdata_pos = 0;

    /* Initialise zlib stream */
    s.zalloc = Z_NULL; /* use default allocation functions */
    s.zfree  = Z_NULL;
    s.opaque = Z_NULL;
    s.next_in  = (unsigned char *)data;
    s.total_in = 0;
    s.next_out  = cdata;
    s.avail_out = cdata_alloc;
    s.total_out = 0;
    s.data_type = Z_BINARY;

    err = deflateInit2(&s, Z_DEFAULT_COMPRESSION, Z_DEFLATED, 15,
		       9, Z_FILTERED);

    /* Encode to 'cdata' array */
    for (i = 0; i < nparts; i++) {
	s.avail_in = part_size[i];
	for (;s.avail_in;) {
	    s.next_out = &cdata[cdata_pos];
	    s.avail_out = cdata_alloc - cdata_pos;
	    if (cdata_alloc - cdata_pos <= 0) {
		fprintf(stderr, "Deflate produced larger output than expected. Abort\n"); 
		return NULL;
	    }
	    err = deflate(&s, Z_SYNC_FLUSH); // also try Z_FULL_FLUSH
	    cdata_pos = cdata_alloc - s.avail_out;
	    if (err != Z_OK) {
		fprintf(stderr, "zlib deflate error: %s\n", s.msg);
		break;
	    }
	}
    }
    if (deflate(&s, Z_FINISH) != Z_STREAM_END) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    *cdata_size = s.total_out;

    if (deflateEnd(&s) != Z_OK) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    return (char *)cdata;
}

static char *zlib_mem_deflate_lparts(char *data,
				     size_t *part_size, int *level, int nparts,
				     size_t *cdata_size) {
    z_stream s;
    unsigned char *cdata = NULL; /* Compressed output */
    int cdata_alloc = 0;
    int cdata_pos = 0;
    int err, i;
    size_t size = 0;

    for (i = 0; i < nparts; i++)
	size += part_size[i];

    cdata = malloc(cdata_alloc = size*1.05+256*nparts);
    cdata_pos = 0;

    /* Initialise zlib stream */
    s.zalloc = Z_NULL; /* use default allocation functions */
    s.zfree  = Z_NULL;
    s.opaque = Z_NULL;
    s.next_in  = (unsigned char *)data;
    s.total_in = 0;
    s.next_out  = cdata;
    s.avail_out = cdata_alloc;
    s.total_out = 0;
    s.data_type = Z_BINARY;

    err = deflateInit2(&s, Z_DEFAULT_COMPRESSION, Z_DEFLATED, 15,
		       9, Z_FILTERED);

    /* Encode to 'cdata' array */
    for (i = 0; i < nparts; i++) {
	s.avail_in = part_size[i];
	for (;s.avail_in;) {
	    s.next_out = &cdata[cdata_pos];
	    s.avail_out = cdata_alloc - cdata_pos;
	    if (cdata_alloc - cdata_pos <= 0) {
		fprintf(stderr, "Deflate produced larger output than expected. Abort\n"); 
		return NULL;
	    }
	    deflateParams(&s, level[i], Z_DEFAULT_STRATEGY);
	    err = deflate(&s, Z_SYNC_FLUSH); // also try Z_FULL_FLUSH
	    cdata_pos = cdata_alloc - s.avail_out;
	    if (err != Z_OK) {
		fprintf(stderr, "zlib deflate error: %s\n", s.msg);
		break;
	    }
	}
    }
    if (deflate(&s, Z_FINISH) != Z_STREAM_END) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    *cdata_size = s.total_out;

    if (deflateEnd(&s) != Z_OK) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    return (char *)cdata;
}

static char *zlib_mem_inflate(char *cdata, size_t csize, size_t *size) {
    z_stream s;
    unsigned char *data = NULL; /* Uncompressed output */
    int data_alloc = 0;
    int err;

    /* Starting point at uncompressed size, 4x compressed */
    if (data_alloc <csize*4+10)
	data = realloc(data, data_alloc = csize*4+10);

    /* Initialise zlib stream */
    s.zalloc = Z_NULL; /* use default allocation functions */
    s.zfree  = Z_NULL;
    s.opaque = Z_NULL;
    s.next_in  = (unsigned char *)cdata;
    s.avail_in = csize;
    s.total_in = 0;
    s.next_out  = data;
    s.avail_out = data_alloc;
    s.total_out = 0;

    err = inflateInit(&s);

    /* Decode to 'data' array */
    for (;s.avail_in;) {
	s.next_out = &data[s.total_out];
	err = inflate(&s, Z_NO_FLUSH);
	if (err == Z_STREAM_END)
	    break;

	if (err != Z_OK) {
	    fprintf(stderr, "zlib inflate error: %s\n", s.msg);
	    break;
	}

	/* More to come, so realloc */
	data = realloc(data, data_alloc += s.avail_in*4 + 10);
	s.avail_out += s.avail_in*4+10;
    }
    inflateEnd(&s);

    *size = s.total_out;
    return (char *)data;
}

static char *mem_deflate(int mode,
			 char *data, size_t size, size_t *cdata_size) {
    switch (mode) {
    case COMP_MODE_NONE:
	return nul_mem_deflate (data, size, cdata_size);
    case COMP_MODE_ZLIB:
	return zlib_mem_deflate(data, size, cdata_size);
#ifdef HAVE_LIBLZMA
    case COMP_MODE_LZMA:
	return lzma_mem_deflate(data, size, cdata_size);
#endif	
    }
    
    return NULL;
}

static char *mem_deflate_parts(int mode, char *data,
			       size_t *part_size, int nparts,
			       size_t *cdata_size) {
    switch (mode) {
    case COMP_MODE_NONE:
	return nul_mem_deflate_parts (data, part_size, nparts, cdata_size);
    case COMP_MODE_ZLIB:
	return zlib_mem_deflate_parts(data, part_size, nparts, cdata_size);
#ifdef HAVE_LIBLZMA
    case COMP_MODE_LZMA:
	return lzma_mem_deflate_parts(data, part_size, nparts, cdata_size);
#endif	
    }
    
    return NULL;
}

static char *mem_deflate_lparts(int mode, char *data,
				size_t *part_size, int *level, int nparts,
				size_t *cdata_size) {
    switch (mode) {
    case COMP_MODE_NONE:
	return nul_mem_deflate_lparts (data, part_size, level, nparts, cdata_size);
    case COMP_MODE_ZLIB:
	return zlib_mem_deflate_lparts(data, part_size, level, nparts, cdata_size);
#ifdef HAVE_LIBLZMA
    case COMP_MODE_LZMA:
	return lzma_mem_deflate_lparts(data, part_size, level, nparts, cdata_size);
#endif	
    }
    
    return NULL;
}

static char *mem_inflate(int mode,
			 char *cdata, size_t csize, size_t *size) {
    switch (mode) {
    case COMP_MODE_NONE:
	return nul_mem_inflate (cdata, csize, size);
    case COMP_MODE_ZLIB:
	return zlib_mem_inflate(cdata, csize, size);
#ifdef HAVE_LIBLZMA
    case COMP_MODE_LZMA:
	return lzma_mem_inflate(cdata, csize, size);
#endif	
    }
    
    return NULL;
}

/*
 * Converts an unsigned integer value into a 7-bit encoded format.
 * We store the bottom 7 bits of value with either 0 or 1 for the top-bit
 * depending on whether any bits are left. We keep repeating this until
 * all significant bits of value have been used.
 *
 * Ie 15551 = hex 3cbf = 0011 1100 1011 1111 becomes:
 *
 *     111 1001   011 1111 (0x3cbf input)
 *    0111 1001  1011 1111 (0x79bf output)
 *
 * Takes an unsigned 32-bit integer and stores in out.
 * Returns the number of bytes written to 'out'
 */
int int2u7(uint32_t in, unsigned char *out) {
    int n = 0;

    out[0] = in & 0x7f;
    while (in >= 128) {
	out[n++] |= 128;
	in >>= 7;
	out[n] = in & 0x7f;
    }

    return n+1;
}

/*
 * Takes a 7-bit encoded value in 'u7' and stores in a
 * 32-bit unsigned int pointed to by 'out'.
 *
 * Returns the number of bytes read from u7.
 */
int u72int(unsigned char *u7, uint32_t *out) {
    uint32_t ret = 0;
    int b = 0;

    ret |= *u7 & 0x7f;
    while (*u7 & 0x80) {
	u7++;
	b += 7;
	ret |= (*u7 & 0x7f) << b;
    }

    *out = ret;
    return b/7+1;
}

/*
 * Converts an signed value into a 7-bit encoded format.
 * We store the bottom 7 bits of value with either 0 or 1 for the top-bit
 * depending on whether any bits are left. We keep repeating this until
 * all significant bits of value have been used.
 *
 * Ie 15551 = hex 3cbf = 0011 1100 1011 1111 becomes:
 *
 *     111 1001   011 1111 (0x3cbf input)
 *    0111 1001  1011 1111 (0x79bf output)
 *
 * Takes an unsigned 32-bit integer and stores in out.
 * Returns the number of bytes written to 'out'
 */
int int2s7(int32_t in, unsigned char *out) {
    int n = 0;
    uint32_t u = (ABS(in) << 1) | (in < 0);

    out[0] = u & 0x7f;
    while (u >= 128) {
	out[n++] |= 128;
	u >>= 7;
	out[n] = u & 0x7f;
    }

    return n+1;
}

/*
 * Takes a 7-bit encoded value in 'u7' and stores in a
 * 32-bit signed int pointed to by 'out'.
 *
 * Returns the number of bytes read from u7.
 */
int s72int(unsigned char *u7, int32_t *out) {
    uint32_t ret = 0;
    int b = 0;

    ret |= *u7 & 0x7f;
    while (*u7 & 0x80) {
	u7++;
	b += 7;
	ret |= (*u7 & 0x7f) << b;
    }

    *out = (ret & 1) ? -(ret >> 1) : (ret >> 1);
    return b/7+1;
}

/* Dynamic growing string */
typedef struct {
    char *str;
    size_t alloc;
    size_t sz;
} dstr;

dstr *dstr_create(void) {
    dstr *ds = malloc(sizeof(*ds));
    if (!ds)
	return NULL;

    ds->str = NULL;
    ds->alloc = 0;
    ds->sz = 0;

    return ds;
}

char *dstr_str(dstr *ds) {
    return ds->str ? ds->str : "";
}

void dstr_empty(dstr *ds) {
    ds->sz = 0;
    if (ds->str)
	*ds->str = 0;
}

void dstr_free(dstr *ds) {
    if (ds) {
	if (ds->str)
	    free(ds->str);
	free(ds);
    }
}

int dstr_append_char(dstr *ds, char c) {
    char *cp;

    if (ds->sz + 2 > ds->alloc) {
	ds->alloc += 8192;
	cp = realloc(ds->str, ds->alloc);
	if (!cp)
	    return -1;
	ds->str = cp;
    }

    cp = &ds->str[ds->sz++];
    *cp++ = c;
    *cp++ = 0;

    return 0;
}

int dstr_append_strn(dstr *ds, char *str, size_t len) {
    char *cp;

    if (ds->sz + len + 1 > ds->alloc) {
	while (ds->sz + len + 1 > ds->alloc) {
	    ds->alloc += 8192;
	}
	cp = realloc(ds->str, ds->alloc);
	if (!cp)
	    return -1;
	ds->str = cp;
    }
    cp = &ds->str[ds->sz];
    memcpy(cp, str, len);
    cp[len] = 0;

    ds->sz += len;

    return 0;
}

int dstr_append_int(dstr *ds, int i) {
    char *cp, *cpo;
    int j, k = 0;

    if (ds->sz + 20 > ds->alloc) {
	ds->alloc += 8192;
	cp = realloc(ds->str, ds->alloc);
	if (!cp)
	    return -1;
	ds->str = cp;
    }

    cpo = cp = &ds->str[ds->sz];

    if (i < 0) {
	*cp++ = '-';
	i = -i;
    } else if (i == 0) {
	*cp++ = '0';
	ds->sz += cp - cpo;
	*cp++ = 0;
	return 0;
    }

    if (i < 1000)
	goto b1;
    if (i < 100000)
	goto b2;
    if (i < 100000000)
	goto b3;

    j = i / 1000000000;
    if (j || k) *cp++ = j + '0', k=1, i %= 1000000000;

    j = i / 100000000;
    if (j || k) *cp++ = j + '0', k=1, i %= 100000000;
    
 b3:
    j = i / 10000000;
    if (j || k) *cp++ = j + '0', k=1, i %= 10000000;
    
    j = i / 1000000;
    if (j || k) *cp++ = j + '0', k=1, i %= 1000000;
    
    j = i / 100000;
    if (j || k) *cp++ = j + '0', k=1, i %= 100000;
    
 b2:
    j = i / 10000;
    if (j || k) *cp++ = j + '0', k=1, i %= 10000;

    j = i / 1000;
    if (j || k) *cp++ = j + '0', k=1, i %= 1000;

 b1:
    j = i / 100;
    if (j || k) *cp++ = j + '0', k=1, i %= 100;

    j = i / 10;
    if (j || k) *cp++ = j + '0', k=1, i %= 10;

    if (i || k) *cp++ = i + '0';

    ds->sz += cp - cpo;
    *cp++ = 0;

    return 0;
}

typedef struct {
    char type[2];
    char format;
    int  len;
    union {
	int32_t i32;
	char *str;
    } val;
} aux_t;

#define MAX_NAME 256
typedef struct sam_entry {
    uint32_t flag, pos, mapq, mpos;
    int32_t isize;
#ifdef SIMPLE_QNAME_MODE
    char *qname;
#else
    char qname[MAX_NAME];
#endif
    char *rname, *cigar, *mrnm, *seq, *qual;
    uint32_t qname_len, rname_len, cigar_len, mrnm_len, seq_len;
#ifdef SIMPLE_READGROUP_MODE
    char *readgroup;
#else
    char readgroup[MAX_NAME];
#endif
    uint32_t readgroup_len;
    aux_t a[50];
    int a_cnt, a_len;
    int delayed;
} sam_entry;

typedef struct {
    sam_entry *sam;
    int idx;
} sam_order;

/* ------------------------------------------------------------------------ */

static int sam_order_cmp(const void *p1, const void *p2) {
    const sam_order *s1 = (const sam_order *)p1;
    const sam_order *s2 = (const sam_order *)p2;
#ifdef SIMPLE_READGROUP_MODE
    if (!s1->sam->readgroup && !s2->sam->readgroup) {
	return s1->sam - s2->sam;
    } else if (!s1->sam->readgroup) {
	return -1;
    } else if (!s2->sam->readgroup) {
	return +1;
    }
#else
    if (!*s1->sam->readgroup && !*s2->sam->readgroup) {
	return s1->sam - s2->sam;
    } else if (!*s1->sam->readgroup) {
	return -1;
    } else if (!*s2->sam->readgroup) {
	return +1;
    }
#endif
    else {
	/*
	int i, j;
	int i_l    = s1->sam->readgroup_len;
	int j_l    = s2->sam->readgroup_len;
	char *i_cp = s1->sam->readgroup;
	char *j_cp = s2->sam->readgroup;

	for (i = j = 0; i < i_l && j < j_l; i++, j++) {
	    if (i_cp[i] < j_cp[i])
		return -1;
	    else if (i_cp[i] > j_cp[i])
		return +1;
	}
	return 0;
	*/

	int d;
	d = strncmp(s1->sam->readgroup, s2->sam->readgroup,
		    MIN(s1->sam->readgroup_len, s2->sam->readgroup_len));
	if (d) {
	    return d;
	} else {
	    d = s1->sam->readgroup_len - s2->sam->readgroup_len;
	    if (d)
		return d;
	    return s1->sam - s2->sam;
	} 
    }
}

typedef struct aux_hist {
    char aux[1024];
    int len;
    struct aux_hist *next;
    struct aux_hist *prev;
} aux_hist_t;

#define NHIST 10

int decomp(FILE *fpi, FILE *fpo, int mode, char *ref, int ref_len) {
    sam_entry *sam;
    sam_order *ord;
    uint32_t count;
    uint32_t csz;
    char *cdata, *data, *cp;
    size_t sz, total_bases;
    int h, i, j, k;
    uint32_t last_pos;
    char *last_str;
    char *seq_region = NULL;
    char *aux_p;

    aux_hist_t aux_hist[NHIST], *aux_hist_h, *aux_hist_f, *aux_hist_t,
	*aux_cp;

    /* Load the next block */
    if (1 != fread(&count, 4, 1, fpi))
	return -1;
    if (1 != fread(&csz, 4, 1, fpi))
	return -1;
    cdata = malloc(csz);
    if (1 != fread(cdata, csz, 1, fpi)) {
	free(cdata);
	return -1;
    }

    switch (mode) {
    case COMP_MODE_ZLIB:
    case COMP_MODE_LZMA:
	cp = data = mem_inflate(mode, cdata, csz, &sz);
	free(cdata);
	break;

    case COMP_MODE_NONE:
	cp = data = cdata;
	sz = csz;
	break;

    default:
	fprintf(stderr, "Unsupported compression mode %d\n", mode);
	return -1;
    }

    fprintf(stderr, "Loaded %d -> %d bytes, count = %d\n",
	   (int)csz, (int)sz, count);

    sam = malloc(count * sizeof(*sam));
    ord = (sam_order *)malloc(count * sizeof(*ord));

    /* read group */
#ifdef SIMPLE_READGROUP_MODE
    for (i = 0; i < count; i++) {
	if (*cp) {
	    sam[i].readgroup = cp;
	    sam[i].readgroup_len = strlen(cp);
	    cp += sam[i].readgroup_len+1;
	} else {
	    sam[i].readgroup = NULL;
	    cp++;
	}
	ord[i].idx = i;
	ord[i].sam = &sam[i];
    }
#else
    {
	char *last = NULL;
	for (i = 0; i < count; i++) {
	    if (*cp != -1) {
		unsigned int j = *cp++, k;
		for (k = 0; k < j; k++) {
		    sam[i].readgroup[k] = last[k];
		}
		do {
		    sam[i].readgroup[k++] = *cp;
		} while (*cp++);
		sam[i].readgroup_len = k-1;
		last = sam[i].readgroup;
	    } else {
		*sam[i].readgroup = 0;
		sam[i].readgroup_len = 0;
		cp++;
	    }

	    ord[i].idx = i;
	    ord[i].sam = &sam[i];
	}
    }
#endif

    /* Sort by read group too so we can optimise later on */
    qsort(ord, count, sizeof(ord[0]), sam_order_cmp);

    /*
    for (i = 0; i < count; i++) {
	printf("%d %d %.*s\n",
	       i, ord[i].idx, ord[i].sam->readgroup_len,
	       ord[i].sam->readgroup);
    }
    */

    /* flags */
    for (i = 0; i < count; i++) {
	cp += u72int((uint8_t *)cp, &sam[i].flag);
    }

    /* Pos */
    last_pos = 1;
    for (i = 0; i < count; i++) {
	cp += u72int((uint8_t *)cp, &sam[i].pos);
	sam[i].pos += last_pos;
	last_pos = sam[i].pos;
    }

    /* MapQ */
    for (i = 0; i < count; i++) {
	cp += u72int((uint8_t *)cp, &sam[i].mapq);
    }

    /* Mate Pos */
    last_pos = 1;
    for (i = 0; i < count; i++) {
	int32_t s32;
	cp += s72int((uint8_t *)cp, &s32);
	if (s32 == 0) {
	    sam[i].mpos = 0;
	} else {
	    sam[i].mpos = (s32-1)/2 + sam[i].pos;
	}
    }
    
    /* Isize */
    for (i = 0; i < count; i++) {
	cp += s72int((uint8_t *)cp, &sam[i].isize);
    }

    /* QName */
#ifdef SIMPLE_QNAME_MODE
    for (i = 0; i < count; i++) {
	ord[i].sam->qname = cp;
	ord[i].sam->qname_len = strlen(cp);
	cp += ord[i].sam->qname_len+1;
    }
#else
    {
	char *last = NULL;
	for (i = 0; i < count; i++) {
	    unsigned int j = *cp++, k;
	    for (k = 0; k < j; k++) {
		ord[i].sam->qname[k] = last[k];
	    }
	    do {
		ord[i].sam->qname[k++] = *cp;
	    } while (*cp++);
	    ord[i].sam->qname_len = k-1;
	    last = ord[i].sam->qname;
	}
    }
#endif

    /* RName */
    last_str = "";
    for (i = 0; i < count; i++) {
	cp += u72int((uint8_t *)cp, &sam[i].rname_len);
	if (sam[i].rname_len == 0) {
	    assert(i != 0);
	    sam[i].rname_len = sam[i-1].rname_len;
	    sam[i].rname = sam[i-1].rname;
	} else {
	    sam[i].rname = cp;
	    cp += sam[i].rname_len;
	    last_str = sam[i].rname;
	}
    }

    /* Cigar */
    for (i = 0; i < count; i++) {
	ord[i].sam->cigar = cp;
	ord[i].sam->cigar_len = strlen(cp);
	cp += ord[i].sam->cigar_len+1;
    }

    /* MRNM */
    for (i = 0; i < count; i++) {
	sam[i].mrnm = cp;
	sam[i].mrnm_len = strlen(cp);
	cp += sam[i].mrnm_len+1;
    }
    
    /* Seq len */
    total_bases = 0;
    for (i = 0; i < count; i++) {
	cp += u72int((uint8_t *)cp, &ord[i].sam->seq_len);
	total_bases += ord[i].sam->seq_len;
    }

    /* Seq */
#ifdef SIMPLE_SEQ_MODE
#if 0
    for (i = 0; i < count; i++) {
	if ((sam[i].flag & 0x04) == 0) {
	    sam[i].seq = cp;
	    cp += sam[i].seq_len;
	}
    }
    for (i = 0; i < count; i++) {
	if ((sam[i].flag & 0x04) != 0) {
	    sam[i].seq = cp;
	    cp += sam[i].seq_len;
	}
    }
#endif

    rle_init();
    for (i = 0; i < count; i++) {
	if ((sam[i].flag & 0x04) == 0) {
	    if (ref) {
		int j, mis = 0, pos;
		int cigar_op;
		int cigar_len;
		char *cigar, *seq;

		sam[i].seq = seq = malloc(sam[i].seq_len+1);

		cigar = sam[i].cigar;
		pos = sam[i].pos-1;

		cigar_len = strtol(cigar, &cigar, 10);
		cigar_op = *cigar++;

		for (j = 0; j < sam[i].seq_len; ) {
		    switch(cigar_op) {
		    case 'S':
		    case 'I':
			//seq[j] = rle_get(&cp);
			//if (seq[j] == ';')
			//    seq[j] = ref[pos];
			j++;
		    break;
		    case 'D':
			pos++;
			break;
		    case 'M':
			//seq[j] = *cp++;
			seq[j] = rle_get(&cp);
			if (seq[j] == ';')
			    seq[j] = ref[pos];
			pos++;
			j++;
			break;
		    default:
			exit(1);
		    }

		    if (j == sam[i].seq_len)
			break;

		    if (--cigar_len == 0) {
			cigar_len = strtol(cigar, &cigar, 10);
			cigar_op = *cigar++;
		    }
		}
		seq[j] = 0;
	    } else {
		sam[i].seq = cp;
		cp += sam[i].seq_len;
	    }
	}
    }
    for (i = 0; i < count; i++) {
	if ((sam[i].flag & 0x04) != 0) {
	    sam[i].seq = cp;
	    cp += sam[i].seq_len;
	} else {
	    if (ref) {
		int j, mis = 0, pos;
		int cigar_op;
		int cigar_len;
		char *cigar, *seq;

		cigar = sam[i].cigar;
		pos = sam[i].pos-1;
		seq = sam[i].seq;

		cigar_len = strtol(cigar, &cigar, 10);
		cigar_op = *cigar++;

		for (j = 0; j < sam[i].seq_len; ) {
		    switch(cigar_op) {
		    case 'S':
		    case 'I':
			seq[j] = *cp++;
			if (seq[j] == ';')
			    seq[j] = ref[pos];
			j++;
		    break;
		    case 'D':
			pos++;
			break;
		    case 'M':
			pos++;
			j++;
			break;
		    default:
			exit(1);
		    }

		    if (j == sam[i].seq_len)
			break;

		    if (--cigar_len == 0) {
			cigar_len = strtol(cigar, &cigar, 10);
			cigar_op = *cigar++;
		    }
		}
	    }
	}
    }
#else
    /*
     * Encode sequence in triplets. This works best is we pad out data
     * to align based on pos%3, so that a sequence shifted by 1 byte
     * still mostly matches the previous encoded string.
     */
    char *cp2 = seq_region = malloc(total_bases);
    if (NULL == cp2)
	return -2;
    for (i = 0; i < count; i++) {
	sam[i].seq = cp2;
	switch (sam[i].pos % 3) {
	case 0:
	    j = 0;
	    break;
	case 1:
	    *cp2++ = "ACGTN"[(*cp /  5) % 5];
	    *cp2++ = "ACGTN"[(*cp     ) % 5];
	    cp++;
	    j = 2;
	    break;
	case 2:
	    *cp2++ = "ACGTN"[(*cp     ) % 5];
	    cp++;
	    j = 1;
	    break;
	}
	for (; j < sam[i].seq_len; j += 3) {
	    *cp2++ = "ACGTN"[(*cp / 25) % 5];
	    *cp2++ = "ACGTN"[(*cp /  5) % 5];
	    *cp2++ = "ACGTN"[(*cp     ) % 5];
	    cp++;
	}
    }
#endif

    /* Qual */
#ifdef SIMPLE_QUAL_MODE
    for (i = 0; i < count; i++) {
	if (*cp == '\377') {
	    ord[i].sam->qual = "*";
	    cp++;
	} else {
	    ord[i].sam->qual = cp;
	    cp += ord[i].sam->seq_len;
	}
    }
#else
    /* Reverse complement - doesn't help that much in the general case */
    for (i = 0; i < count; i++) {
	if (*cp == '\377') {
	    ord[i].sam->qual = "*";
	    cp++;
	} else {
	    ord[i].sam->qual = cp;
	    cp += ord[i].sam->seq_len;
	    if ((ord[i].sam->flag & 0x10) != 0) {
		char *q = ord[i].sam->qual;
		for (j = 0, k = ord[i].sam->seq_len-1; j < k; j++, k--) {
		    char t = q[j];
		    q[j] = q[k];
		    q[k] = t;
		}
	    }
	}
    }
#endif

    /* Aux */
#ifdef SIMPLE_AUX_MODE
    for (i = 0; i < count; i++) {
	int data_len = 0;
	size_t l;

	sam[i].a_cnt = *cp++;
	for (j = 0; j < sam[i].a_cnt; j++) {
	    sam[i].a[j].type[0] = *cp++;
	    sam[i].a[j].type[1] = *cp++;

	    if (sam[i].a[j].type[0] == 'R' &&
		sam[i].a[j].type[1] == 'G') {
		sam[i].a[j].format = 'Z';
		sam[i].a[j].val.str = sam[i].readgroup;
		continue;
	    }

	    sam[i].a[j].format  = *cp++;
	    switch(sam[i].a[j].format) {
	    case '0':
		sam[i].a[j].val.i32 = 0;
		data_len++;
		break;

	    case 'i':
		cp += u72int((uint8_t *)cp, &sam[i].a[j].val.i32);
		data_len += 12;
		break;

	    case 'I':
		cp += u72int((uint8_t *)cp, &sam[i].a[j].val.i32);
		sam[i].a[j].val.i32 =- sam[i].a[j].val.i32;
		data_len += 12;
		break;

	    case 'A':
		sam[i].a[j].val.str = cp;
		cp++;
		data_len++;
		break;

	    case 'Z':
		sam[i].a[j].val.str = cp;
		l = strlen(cp);
		data_len += l;
		cp += l + 1;
		break;

	    default:
		fprintf(stderr, "Unknown aux format %c\n",
			sam[i].a[j].format);
	    }
	}
	sam[i].a_len = j*6 /* \t + type[2] + colon + fmt + colon */ + data_len;
    }
#else

    /*
     * Track recent aux formats and output symbols to represent
     * "2nd most recent format string" instead of the actual formats.
     * Currently we track up to 3 formats, but we may need to implement
     * more.
     * FIXME: Use pointers instead of memcpy too.
     */
    for (h = 0; h < NHIST; h++) {
	aux_hist[h].len = 0;
	aux_hist[h].prev = h ? &aux_hist[h-1] : NULL;
	aux_hist[h].next = h+1 < NHIST ? &aux_hist[h+1] : NULL;
    }
    aux_hist_h = NULL;         /* used list head */
    aux_hist_t = NULL;         /* used list tail */
    aux_hist_f = &aux_hist[0]; /* free list */

    for (i = 0; i < count; i++) {
	int data_len;
	size_t l, aux_len;
	//sam_entry *samp = &sam[i];
	sam_entry *samp = ord[i].sam;

	samp->delayed = -1;	

	/* Get the encoding pattern */
	int vers;
	vers = *cp++;

	//printf("h=%d ", vers);

	if (vers == 0) {
	    /* Assumption, no nul chars in format codes */
	    aux_p = cp;
	    while (*cp++);
	    aux_len = cp - aux_p;

	    /* Insert new aux_p into aux_hist array */
	    if (aux_hist_f) {
		aux_cp = aux_hist_f;
		aux_hist_f = aux_hist_f->next;
	    } else {
		aux_cp = aux_hist_t;
		aux_hist_t->prev->next = NULL;
		aux_hist_t = aux_hist_t->prev;
	    }

	    aux_cp->next = aux_hist_h;
	    aux_cp->prev = NULL;
	    if (aux_hist_h)
		aux_hist_h->prev = aux_cp;
	    aux_hist_h = aux_cp;
	    if (!aux_hist_t)
		aux_hist_t = aux_cp;

	    aux_cp->len = aux_len;
	    memcpy(aux_cp->aux, aux_p, aux_len);
	    aux_cp->aux[aux_len] = 0;

	} else {
	    if (vers > NHIST) {
		fprintf(stderr, "Unexpected aux history version %d\n",
			vers);
		abort();
	    }

	    aux_cp = aux_hist_h;
	    while (--vers)
		aux_cp = aux_cp->next;

	    /* Move aux_cp to start of list */
	    if (aux_cp != aux_hist_h) {
		if (aux_hist_t == aux_cp)
		    aux_hist_t = aux_cp->prev;
		if (aux_cp->prev)
		    aux_cp->prev->next = aux_cp->next;
		if (aux_cp->next)
		    aux_cp->next->prev = aux_cp->prev;

		aux_cp->next = aux_hist_h;
		aux_cp->prev = NULL;
		aux_hist_h->prev = aux_cp;
		aux_hist_h = aux_cp;
	    }

	    aux_p = aux_cp->aux;
	    aux_len = aux_cp->len;
	}   

	//printf("(%.*s)\n", aux_len, aux_p);

	/* Decode */
	j = data_len = 0;
	while (*aux_p) {
	    samp->a[j].type[0] = *aux_p++;
	    samp->a[j].type[1] = *aux_p++;

	    /* Read group has already been decoded */
	    if (samp->a[j].type[0] == 'R' &&
		samp->a[j].type[1] == 'G') {
		samp->a[j].format = 'Z';
		samp->a[j].val.str = samp->readgroup;
		data_len += samp->readgroup_len;
		j++;
		continue;
	    }

#if 1
	    /* Delay OQ until after the other aux records */
	    if (samp->a[j].type[0] == 'O' &&
		samp->a[j].type[1] == 'Q') {
		samp->delayed = j;
		aux_p++;
		j++;
		continue;
	    }
#endif

	    switch (*aux_p++) {
	    case '0':
		samp->a[j].format  = 'i';
		samp->a[j].val.i32 = 0;
		data_len++;
		break;

	    case 'i':
		samp->a[j].format  = 'i';
		cp += u72int((uint8_t *)cp,
			     (unsigned int *)&samp->a[j].val.i32);
		data_len += 12;
		break;

	    case 'I':
		samp->a[j].format  = 'i';
		cp += u72int((uint8_t *)cp,
			     (unsigned int *)&samp->a[j].val.i32);
		samp->a[j].val.i32 = -samp->a[j].val.i32;
		data_len += 12;
		break;

	    case 'A':
		samp->a[j].format  = 'A';
		samp->a[j].val.str = cp;
		data_len++;
		cp++;
		break;

	    case 'Z':
		samp->a[j].format  = 'Z';
		samp->a[j].val.str = cp;
		l = strlen(cp);
		data_len += l;
		cp += l + 1;
		break;

	    default:
		fprintf(stdout, "Unknown aux format %c\n", cp[-1]);
	    }

	    j++;
	}
	samp->a_cnt = j;
	samp->a_len = j*6 /* \t + type[2] + colon + fmt + colon */ + data_len;
    }

#if 1
    /* Load delays OQ:Z: records */
    for (i = 0; i < count; i++) {
	sam_entry *samp = ord[i].sam;
	size_t l;

	if ((j = samp->delayed) == -1)
	    continue;

	samp->a[j].format = 'Z';
	samp->a[j].val.str = cp;
	l = strlen(cp);
	cp += l + 1;
	samp->a_len += l;
    }
#endif
#endif

    /* ------ */
    
    /* Dump out the results */
    {
	size_t aux_len = 100;
	char *aux_str = malloc(aux_len);

	dstr *ds = dstr_create();

	for (i = 0; i < count; i++) {
	    /*
	     * Or via a fast fprintf:
	     */
	    /*
	      faster_fprintf(fpo,
	          "%.*s\t%d\t%.*s\t%d\t%d\t%.*s\t%.*s\t%d\t%d\t%.*s\t%.*s",
		  sam[i].qname_len, sam[i].qname,
		  sam[i].flag,
		  sam[i].rname_len, sam[i].rname,
		  sam[i].pos,
		  sam[i].mapq,
		  sam[i].cigar_len, sam[i].cigar,
		  sam[i].mrnm_len,  sam[i].mrnm,
		  sam[i].mpos,
		  sam[i].isize,
		  sam[i].seq_len,   sam[i].seq,
		  sam[i].seq_len,   sam[i].qual);
	    */

	    dstr_empty(ds);
	    dstr_append_strn(ds, sam[i].qname, sam[i].qname_len);

	    dstr_append_char(ds, '\t');
	    dstr_append_int(ds, sam[i].flag);

	    dstr_append_char(ds, '\t');
	    dstr_append_strn(ds, sam[i].rname, sam[i].rname_len);

	    dstr_append_char(ds, '\t');
	    dstr_append_int(ds, sam[i].pos);

	    dstr_append_char(ds, '\t');
	    dstr_append_int(ds, sam[i].mapq);

	    dstr_append_char(ds, '\t');
	    dstr_append_strn(ds, sam[i].cigar, sam[i].cigar_len);

	    dstr_append_char(ds, '\t');
	    dstr_append_strn(ds, sam[i].mrnm, sam[i].mrnm_len);

	    dstr_append_char(ds, '\t');
	    dstr_append_int(ds, sam[i].mpos);

	    dstr_append_char(ds, '\t');
	    dstr_append_int(ds, sam[i].isize);

	    dstr_append_char(ds, '\t');
	    dstr_append_strn(ds, sam[i].seq, sam[i].seq_len);

	    dstr_append_char(ds, '\t');
	    if (sam[i].qual[0] == '*' && sam[i].qual[1] == 0) {
		dstr_append_char(ds, '*');
	    } else {
		dstr_append_strn(ds, sam[i].qual, sam[i].seq_len);
	    }

	    fwrite(ds->str, 1, ds->sz, fpo);


	    if (aux_len <= sam[i].a_len) {
		aux_len = sam[i].a_len+1;
		aux_str = realloc(aux_str, aux_len);
	    }
	    cp = aux_str;

	    for (j = 0; j < sam[i].a_cnt; j++) {
		*cp++ = '\t';
		*cp++ = sam[i].a[j].type[0];
		*cp++ = sam[i].a[j].type[1];
		*cp++ = ':';

		switch(sam[i].a[j].format) {
		case '0':
		    *cp++ = 'i';
		    *cp++ = ':';
		    *cp++ = '0';
		    break;

		case 'i':
		case 'I': {
		    char ibuf[20];
		    int k, v;

		    *cp++ = 'i';
		    *cp++ = ':';
		
		    v = sam[i].a[j].val.i32;
		    if (v) {
			if (v < 0) {
			    *cp++ = '-';
			    v = -v;
			}
			for (k = 0; v; v /= 10) {
			    ibuf[k++] = v % 10;
			}
			while (--k >= 0) {
			    *cp++ = ibuf[k] + '0';
			}
		    } else {
			*cp++ = '0';
		    }

		    break;
		}

		case 'A':
		    *cp++ = 'A';
		    *cp++ = ':';
		    *cp++ = *sam[i].a[j].val.str;
		    break;

		case 'Z': {
		    char *b = sam[i].a[j].val.str;

		    *cp++ = 'Z';
		    *cp++ = ':';
		    while ((*cp++ = *b++))
			;
		    cp--;

		    break;
		}
		}
	    }
	    *cp++ = '\n';
	    fwrite(aux_str, 1, cp-aux_str, fpo);

	    if (ref && (sam[i].flag & 0x04) == 0)
		free(sam[i].seq);
	}

	free(aux_str);

	dstr_free(ds);
    }

    free(sam);
    free(data);
    if (seq_region)
	free(seq_region);
    free(ord);

    return 0;
}

static unsigned char *load(int *lenp, char *fn) {
    unsigned char *data = NULL;
    int dsize = 0;
    int dcurr = 0, len;
    int fd = 0;

    if (fn) {
	if (-1 == (fd = open(fn, O_RDONLY, 0))) {
	    perror(fn);
	    return NULL;
	}
    }

    do {
	if (dsize - dcurr < 8192) {
	    dsize = dsize ? dsize * 2 : 8192;
	    if (NULL == (data = realloc(data, dsize))) {
		if (fd)
		    close(fd);
		return NULL;
	    }
	}

	len = read(fd, data + dcurr, 8192);
	if (len > 0)
	    dcurr += len;
    } while (len > 0);

    if (len == -1) {
	perror("read");
	if (fd)
	    close(fd);
	return NULL;
    }

    if (fd)
	close(fd);

    *lenp = dcurr;
    return data;
}

#define IO_BUF_SIZE 2*1024*1024

int main(int argc, char **argv) {
    char hdr[16];
    int mode;
    int32_t csz;
    size_t usz;
    char *cdata, *udata;
    FILE *in = stdin, *out = stdout;
    char *in_buf  = malloc(IO_BUF_SIZE);
    char *out_buf = malloc(IO_BUF_SIZE);
    char *ref = NULL;
    int ref_len = 0;

    if (argc > 2 && strcmp(argv[1], "-r") == 0) {
	ref = load(&ref_len, argv[2]);
	argc-=2;
	argv+=2;
    }

    /*
     * If we have GNU malloc then disable the use of mmap and tidying up
     * of freed blocks. It's pointless as our memory is constant usage
     * and zlib will immediately malloc again.
     */
    mallopt(M_MMAP_MAX, 0);
    mallopt(M_TRIM_THRESHOLD, -1);

    if (argc >= 2) {
	if (NULL == (in = fopen(argv[1], "rb"))) {
	    perror(argv[1]);
	    exit(1);
	}
    }
    if (argc >= 3) {
	if (NULL == (out = fopen(argv[2], "wb"))) {
	    perror(argv[2]);
	    exit(1);
	}
    }

    setvbuf(in,  in_buf,  _IOFBF, IO_BUF_SIZE);
    setvbuf(out, out_buf, _IOFBF, IO_BUF_SIZE);

    /* Decode ZAM header */
    if (16 != fread(hdr, 1, 16, in))
	return 1;

    if (memcmp(hdr, ".zam1.00", 8) != 0)
	return 1;
    mode = hdr[8];

    /* SAM header */
    if (4 != fread(&csz, 1, 4, in))
	return 1;
    cdata = malloc(csz);
    if (csz != fread(cdata, 1, csz, in))
	return 1;

    switch (mode) {
    case COMP_MODE_ZLIB:
    case COMP_MODE_LZMA:
	udata = mem_inflate(mode, cdata, csz, &usz);
	free(cdata);
	break;

    case COMP_MODE_NONE:
	udata = cdata;
	usz = csz;
	break;

    default:
	return 1;
    }
    fwrite(udata, 1, usz, out);
    free(udata);

    /* ZAM blocks */
    while (0 == decomp(in, out, mode, ref, ref_len));
    //decomp(in, out, mode);

    if (in != stdin)
	fclose(in);
    if (out != stdout)
	fclose(out);

    fflush(stdout);

    free(in_buf);
    free(out_buf);
    
    return 0;
}
